#include <stdlib.h>
#include <ctype.h>
#include <ncurses.h>

#include "Engine.h"
#include "Snake.h"
#include "Food.h"

/* * *
 * Create a board with border with height according to the maximum values 
 * of the standard screen.
 * * */
void Board_Create(Board* board) {
	unsigned int height, width;
	// Get the size of standard screen.
	unsigned int yMax, xMax;
	getmaxyx(stdscr, yMax, xMax);
	height = yMax/1.5;
	width  = xMax/1.5;
	// Create a window (boardWin) for snake to run in.
	WINDOW* boardWin = newwin(height, width, 
			(yMax/2)-(height/2), (xMax/2)-(width/2));
	// Initialize value to the board passed in.
	board->win = boardWin;
	getbegyx(boardWin, board->yLoc, board->xLoc);
	board->yMax = yMax;
	board->xMax = xMax;
	board->height = height;
	board->width = width;
}

/* * *
 * Draw the snake and the food inside the board passed in.
 * * */
void Engine_Draw(Board* board, Snake_t* snake, Food_t* food, const int* level, const unsigned short* score, unsigned int* snakeLen) {
	// Turn on attribute to draw snake and food.
	wattron(board->win, A_STANDOUT);
	// Draw the snake.
	Snake_Draw(board, &snake);
	// Draw the food.
	Food_Draw(board, food);
	// Turn off attribute.
	wattroff(board->win, A_STANDOUT);
	// Print the level statistic.
	mvprintw(board->yLoc - 1, board->xLoc + 0, 
			"| Speed: %03d ms | Score: %02u | Length: %02d |"
			, *level, *score, *snakeLen);
	refresh();
}

/* * *
 * <<< SNAKE GAME ENGINE >>>
 * Run snake starting with snake_length inside board passed in.
 * Change snake direction according to user's input.
 * Extend snake_length by one when snake eat one food.
 * Stop when snake have collision or when snake is out of bound.
 * * */
void Engine_Run(Board* board, Snake_t* snake, unsigned int* snake_length) {
	/* <<< FLOW CONTROL VARS >>> */
	bool gameEnd = false;
	Food_t* food = NULL;
	bool noFood  = true;
	int input    = KEY_RIGHT;
	int oldInput = 0;
	int level    = 400;
	unsigned short score = 1;
	Snake_t* prob = NULL;
	// Print user's control hint.
	mvprintw(board->yLoc + board->height, board->xLoc + board->width - 35, 
			"[Space] to pause | [End] to quit."
			);
	/* <<< RUN SNAKE >>> */
	// Enable keyboard's input.
	keypad(board->win, true);
	while (gameEnd == false) {
		// Set pausing time.
		wtimeout(board->win, level);
		// Draw all elements.
		Engine_Draw(board, snake, food, &level, &score, snake_length);
		// Check for no food - create food.
		if (noFood == true) {
			food = Food_Init(board);
			// Check for memor error - throw error and end the game.
			if (food == NULL) {
				mvwprintw(stdscr, board->yLoc - 1, board->xLoc + 2,
						"Could not allocate memory for food");
				wrefresh(stdscr);
				getch();
				gameEnd = true;
			}
			noFood = false;
		}
		// Set old input to current input.
		oldInput = input;
		// Get input from user.
		input = wgetch(board->win);
		// Check for no input - reset back to old input.
		if (input == ERR) input = oldInput;
		// Erase the snake.
		Snake_Erase(board, &snake);
		// Parse input and move the snake.
		// Check for reverse - reset input and no more action.
		switch (tolower(input)) {
			case KEY_LEFT: // Left.
				if (snake->yLoc == snake->next->yLoc 
						&& snake->xLoc == snake->next->xLoc + 1) {
					input = oldInput; 
					break;
				}
				// Change the snake's coordinate.
				Snake_Move_Left(&snake);
				break;
			case KEY_DOWN: // Down.
				if (snake->yLoc == snake->next->yLoc - 1 
						&& snake->xLoc == snake->next->xLoc) {
					input = oldInput; 
					break;
				}
				// Change the snake's coordinate.
				Snake_Move_Down(&snake);
				break;
			case KEY_UP: // Up.
				if (snake->yLoc == snake->next->yLoc + 1 
						&& snake->xLoc == snake->next->xLoc) {
					input = oldInput;
				   	break;
				}
				// Change the snake's coordinate.
				Snake_Move_Up(&snake);
				break;
			case KEY_RIGHT: // Right.
				if (snake->yLoc == snake->next->yLoc 
						&& snake->xLoc == snake->next->xLoc - 1) {
					input = oldInput;
				   	break;
				}
				// Change the snake's coordinate.
				Snake_Move_Right(&snake);
				break;
			case KEY_END: // Quit.
				gameEnd = true;
				break;
			case ' ': // Pause.
				cbreak();
				input = KEY_EOL;
				// Set pausing time.
				wtimeout(board->win, level);
				break;
			default:
				input = oldInput;
				break;
		}
		// Check for collision - end game.
		prob = snake->next;
		while (prob != NULL) {
			// Check head coordinate match with probe.
			if (snake->yLoc == prob->yLoc && snake->xLoc == prob->xLoc) {
				gameEnd = true;
			}
			// Move probe to next node.
			prob = prob->next;
		} prob = NULL;
		// Check for eating - extend snake and increase level.
		if (snake->yLoc == food->yLoc && snake->xLoc == food->xLoc) {
			// Extend the snake according to value of food.
			for (int i = 0; i < food->vGrow; i++) {
				Snake_Add(&snake, food->yLoc, food->xLoc);
			}
			// Increase score and snake's length.
			score++; (*snake_length) += food->vGrow;
			// Increase difficulty level.
			if (level > 200) level -= 50;
			if (level > 100) level -= 10;
			if (level > 50) level  -= 1;
			if (level > 40) level  -= 1;
			if (level == 1) level  = 1;
			// De-allocate memory of food.
			free(food);
			// Flag to create more food.
			noFood = true;
		}
		// Check for out of bound - end game.
		if  (  
				snake->yLoc == 0 ||						// Top bound.
				snake->yLoc == board->height - 1 ||		// Bottom bound.
				snake->xLoc == board->width - 1 ||		// Right bound.
				snake->xLoc == 0						// Left bound.
			) {
			gameEnd = true;
		}
	}
	// Erase and free any food left before exit.
	if (food != NULL) {
		mvwprintw(board->win, food->yLoc, food->xLoc, " ");
		free(food);
	}
}
